/* 
** Unmount the davfs file system.
 */

/*
 *  This file is part of davfs2.
 *
 *  davfs2 is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  davfs2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with davfs2; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */


#include "config.h"

#include <error.h>
#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <ne_string.h>

#include "defaults.h"


/* This is lazy programming. All the dirty work is left to the real umount
   program, while we just sit and wait for mount.davfs to terminate.
   umount.davfs is a umount helper. It is usually called by umount and makes
   sure, that umount will not return until mount.davfs has synchronized all
   files.
   It first reads the pid-file and identifies the mount.davfs process. Then
   it calls mount again, with option -i (to not be called again), to do the
   real unmounting. In a loop it will watch the process list. When the
   mount.davfs process terminates, it will return.
   If it can't identify the mount.davfs process, it will call umount -i anyway,
   but warn the user. */
int main(int argc, char *argv[]) {

    setuid(getuid());
    setgid(getgid());
    
    static const struct option options[] = { {0, 0, 0, 0} };
    while (getopt_long(argc, argv, "", options, NULL) != -1);

    if (optind != (argc - 1)) {
        printf("%s: Can't read command line.\n", argv[0]);
        exit(EXIT_FAILURE);
    }
    char *mpoint = canonicalize_file_name(argv[optind]);
    if (mpoint == NULL)
        mpoint = argv[optind];
    if (mpoint == NULL || *mpoint != '/') {
        printf("%s: Can't read mount point.\n", argv[0]);
        exit(EXIT_FAILURE);
    }

    char *m = mpoint;
    while (*m == '/')
        m++;
    char *mp = ne_strdup(m);
    m = strchr(mp, '/');
    while (m != NULL) {
        *m = '-';
        m = strchr(mp, '/');
    }
    char *pidfile = ne_concat(DAV_SYS_RUN, "/", mp, ".pid", NULL);
    free(mp);

    char *umount_command = ne_concat("umount -i ", mpoint, NULL);

    char *pid = NULL;
    FILE *file = fopen(pidfile, "r");
    if (file == NULL || fscanf(file, "%a[0-9]", &pid) != 1
            || pid == NULL) {
        printf("%s:\n"
               "    Can't read PID from file \"%s\".\n"
               "    I am trying to unmount.\n"
               "    Please wait for %s to terminate (using \"ps x\").\n",
               argv[0], pidfile, PROGRAM_NAME);
        return system(umount_command);
    }
    fclose(file);

    char *ps_command = ne_concat("ps -p ", pid, " -C ", mpoint, NULL);
    FILE *ps_in = popen(ps_command, "r");
    if (ps_in == NULL) {
        printf("%s:\n"
               "    Can't read process list.\n"
               "    I am trying to unmount.\n"
               "    Please wait for %s to terminate (using \"ps x\").\n",
               argv[0], PROGRAM_NAME);
        return system(umount_command);
    }

    int found = 0;
    size_t n = 0;
    char *ps_line = NULL;
    while (!found && getline(&ps_line, &n, ps_in) > 0)
        found = (strstr(ps_line, pid) != NULL
                 && strstr(ps_line, PROGRAM_NAME) != NULL);

    pclose(ps_in);

    if (!found) {
        printf("%s:\n"
               "    Can't find %s-process with pid %s.\n"
               "    I am trying to unmount anyway.\n"
               "    You propably have to remove %s manually.\n",
               argv[0], PROGRAM_NAME, pid, pidfile);
        return system(umount_command);
    }

    if (system(umount_command) != 0)
        exit(EXIT_FAILURE);

    printf("%s: Waiting while %s (pid %s) synchronizes the cache .",
           argv[0], PROGRAM_NAME, pid);
    fflush(stdout);

    while (found) {

        sleep(3);
        printf(".");
        fflush(stdout);

        ps_in = popen(ps_command, "r");
        if (ps_in == NULL) {
            printf("\n%s: An error occurred while waiting.\n"
                   "    Please using \"ps x\" to wait for %s to terminate.\n",
                   argv[0], PROGRAM_NAME);
            exit(EXIT_FAILURE);
        }

        found = 0;
        while (!found && getline(&ps_line, &n, ps_in) > 0)
            found = (strstr(ps_line, pid) != NULL
                     && strstr(ps_line, PROGRAM_NAME) != NULL);

        pclose(ps_in);
    }
    printf("OK\n");

    return 0;
}
